/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.anastau;

import java.text.MessageFormat;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.TimeUnit;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.comm.GanglinienPrognoseClient;
import de.bsvrz.iav.gllib.gllib.portals.DavGanglinienprognose;
import de.bsvrz.iav.gllib.gllib.portals.DavObjektFactory;
import de.bsvrz.ibv.anastau.bestimmungstau.IStauobjektBestimmung;
import de.bsvrz.ibv.anastau.bestimmungstau.StauobjektBestimmung;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.IStau;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.StauDefs;
import de.bsvrz.ibv.anastau.domain.SvaBetriebsmeldungssender;
import de.bsvrz.ibv.anastau.migration.daf.Konfigurationsbereich;
import de.bsvrz.ibv.anastau.migration.modell.ObjektFactory;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.VerkehrsModellTypen;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.ibv.anastau.portale.DavSvaBetriebsmeldungssender;
import de.bsvrz.ibv.anastau.progstau.IStauverlaufsPrognose;
import de.bsvrz.ibv.anastau.progstau.StauverlaufsPrognose;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Hauptklasse der SWE - Stauverlaufsanalyse.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public final class StauverlaufsAnalyse extends TimerTask
        implements StandardApplication, Thread.UncaughtExceptionHandler {

    /**
     * Logger für Ausgabe von Meldungen.
     */
    private static Debug logger;

    /**
     * Kommandozeilenparameter für die Angabe der Klasse, die als
     * Stauobjektbestimmungsmodul verwendet werden soll.
     */
    private static final String PARAMETER_SOB_MODUL = "-sobModul";

    /**
     * Kommandozeilenparameter für die Angabe der Klasse, die als
     * Stauverlaufsprognosemodul verwendet werden soll.
     */
    private static final String PARAMETER_PROGNOSE_MODUL = "-prognoseModul";

    /**
     * Kommandozeilenparameter für die Angabe des VerkehrsmodellNetz, für das die
     * Staus ermittelt und verwaltet werden sollen.
     */
    private static final String PARAMETER_NETZ = "-netz";

    /**
     * Standardklasse für die Implementierung der Stauobjektbestimmung.
     */
    private static final String DEFAULT_SOB_MODUL = StauobjektBestimmung.class.getName();

    /**
     * Standardklasse für die Implementierung der Stauverlaufsanalyse.
     */
    private static final String DEFAULT_PROGNOSE_MODUL = StauverlaufsPrognose.class.getName();

    /**
     * die verwendete Datenverteilerverbindung.
     */
    private ClientDavInterface dav;

    /**
     * die globalen Optionen der Stauverlaufsanalyse.
     */
    private static StauverlaufsAnalyseOptionen optionen;

    /**
     * liefert die Datenverteilerverbindung der Anwendung.
     *
     * @return die Verbindung
     */
    public ClientDavInterface getDav() {
        return dav;
    }

    /**
     * liefert fir globalen Optionen der Stauverlaufsanalyse.
     *
     * @return die Optionen
     */
    public static StauverlaufsAnalyseOptionen getOptionen() {
        return StauverlaufsAnalyse.optionen;
    }

    /**
     * Hauptfunktion der Anwendung.
     *
     * @param args die Liste der &uuml;bergebenen Kommandozeilenparameter
     */
    public static void main(final String[] args) {
        StandardApplicationRunner.run(new StauverlaufsAnalyse(), args);
    }

    /**
     * der Haupttimer zur zyklischen Ausführung der Stauobjektbstimmung und der
     * nachfolgenden Prognose.
     */
    private final Timer mainTimer = new Timer();

    /**
     * die Liste der Module zur Stauobjektbestimmung. Für jedes VerkehrsmodellNetz
     * wird ein eigenes Modul gestartet.
     */
    private final Map<VerkehrModellNetz, IStauobjektBestimmung> sobModule = new HashMap<>();

    /**
     * der Name des Moduls für die Stauobjektbestimmung.
     */
    private String sobName;

    /**
     * der Name des Moduls für die Stauverlaufsprognose.
     */
    private String prognoseName;

    /**
     * die Liste der PIDs der Netze, die verwaltet werden sollen.
     */
    private final Set<String> netzPids = new HashSet<>();

    /**
     * die Netze die bearbeitet werden.
     */
    private final Set<VerkehrModellNetz> netzObjekte = new HashSet<>();

    private GanglinienPrognoseClient prognose;

    private SvaBetriebsmeldungssender betriebsmeldungssender;

	private int maxGapSize;

    /**
     * Standardkonstruktor.
     */
    private StauverlaufsAnalyse() {
        super();
        ObjektFactory.getInstanz().registerStandardFactories();
    }

    @Override
    public void initialize(final ClientDavInterface connection) throws Exception {
        dav = connection;
        betriebsmeldungssender = new DavSvaBetriebsmeldungssender(connection);
        prognose = new DavGanglinienprognose(new DavObjektFactory(dav));

        MessageSender.getInstance()
                .setApplicationLabel(getClass().getSimpleName() + "_" + connection.getLocalApplicationObject().getId());

        ObjektFactory.getInstanz().setVerbindung(connection);

        /* Alte Staus entfernen */
        for (final SystemObject obj : connection.getDataModel().getType(StauDefs.PID_TYP_STAU).getElements()) {
            obj.invalidate();
        }

        if (netzPids.isEmpty()) {
            StauverlaufsAnalyse.logger.config("Keine Netze vorgegeben, es werden alle konfigurierten Netze verwaltet");
            for (final SystemObject objekt : connection.getDataModel()
                    .getType(VerkehrsModellTypen.VERKEHRSMODELLNETZ.getPid()).getElements()) {
                StauverlaufsAnalyse.logger.config("Lade Netzobjekt: " + objekt);
                VerkehrModellNetz netz = (VerkehrModellNetz) ObjektFactory.getInstanz().getModellobjekt(objekt);
                if (netz.isSchreibbar(connection.getLocalConfigurationAuthority())) {
                    netzObjekte.add(netz);
                }
            }
        } else {
            for (final String pid : netzPids) {
                final SystemObject pidObj = connection.getDataModel().getObject(pid);
                if (pidObj instanceof ConfigurationArea) {
                    StauverlaufsAnalyse.logger.config(
                            "Initialisiere alle Netze des Konfigurationsbereichs: " + pidObj.getNameOrPidOrId());
                    for (final SystemObject caObj : Konfigurationsbereich.getObjekte((ConfigurationArea) pidObj,
                            VerkehrsModellTypen.VERKEHRSMODELLNETZ.getPid())) {
                        VerkehrModellNetz netz = (VerkehrModellNetz) ObjektFactory.getInstanz().getModellobjekt(caObj);
                        if (netz.isSchreibbar(connection.getLocalConfigurationAuthority())) {
                            netzObjekte.add(netz);
                        }
                    }
                } else {
                    StauverlaufsAnalyse.logger.config("Initialisiere das Netz: " + pidObj.getNameOrPidOrId());
                    VerkehrModellNetz netz = (VerkehrModellNetz) ObjektFactory.getInstanz().getModellobjekt(pidObj);
                    if (netz.isSchreibbar(connection.getLocalConfigurationAuthority())) {
                        netzObjekte.add(netz);
                    }
                }
            }
        }

        if (netzObjekte.isEmpty()) {
            StauverlaufsAnalyse.logger.error("Es sind keine gültigen Objekte vom Typ VerkehrsmodellNetz verfügbar!");
            System.exit(-1);
        }

        final Class<?> sobClass = Class.forName(sobName);
        final Class<?> prognoseClass = Class.forName(prognoseName);

        for (final VerkehrModellNetz netzObjekt : netzObjekte) {
            StauverlaufsAnalyse.logger.config("Initialisiere Module für Netzobjekt: " + netzObjekt.getName());

            StauverlaufsAnalyse.logger.config("Initialisiere Staubestimmungsmodul: " + sobClass.getName());
            final IStauobjektBestimmung sobModul = (IStauobjektBestimmung) sobClass.getConstructor(new Class[] {})
                    .newInstance(new Object[] {});

            StauverlaufsAnalyse.logger.config("Initialisiere Stauprognosemodul: " + sobClass.getName());
            final IStauverlaufsPrognose prognoseModul = (IStauverlaufsPrognose) prognoseClass
                    .getConstructor(new Class[] {}).newInstance(new Object[] {});
            prognoseModul.init(this, netzObjekt);

            sobModul.init(this, netzObjekt, prognoseModul);
            sobModule.put(netzObjekt, sobModul);
        }

        mainTimer.scheduleAtFixedRate(this, 0, TimeUnit.SECONDS.toMillis(1));

        StauverlaufsAnalyse.logger.config("Alle Module der Stauverlaufsprognose wurden initialisiert");
    }

    @Override
    public void parseArguments(final ArgumentList argList) throws Exception {

        StauverlaufsAnalyse.logger = Debug.getLogger();
        Thread.setDefaultUncaughtExceptionHandler(this);

        if (argList.hasArgument(StauverlaufsAnalyse.PARAMETER_SOB_MODUL)) {
            sobName = argList.fetchArgument(StauverlaufsAnalyse.PARAMETER_SOB_MODUL).asString();
        }
        if (argList.hasArgument(StauverlaufsAnalyse.PARAMETER_PROGNOSE_MODUL)) {
            prognoseName = argList.fetchArgument(StauverlaufsAnalyse.PARAMETER_PROGNOSE_MODUL).asString();
        }
        if (argList.hasArgument(StauverlaufsAnalyse.PARAMETER_NETZ)) {
            final String netzNamen = argList.fetchArgument(StauverlaufsAnalyse.PARAMETER_NETZ).asString();
            final StringTokenizer tokenizer = new StringTokenizer(netzNamen, ",");
            while (tokenizer.hasMoreElements()) {
                netzPids.add(tokenizer.nextToken().trim());
            }
        }

        maxGapSize = argList.fetchArgument("-maxGapSize=0").intValue();
        
        if ((sobName == null) || (sobName.length() <= 0)) {
            sobName = StauverlaufsAnalyse.DEFAULT_SOB_MODUL;
        }
        if ((prognoseName == null) || (prognoseName.length() <= 0)) {
            prognoseName = StauverlaufsAnalyse.DEFAULT_PROGNOSE_MODUL;
        }

        StauverlaufsAnalyse.optionen = new StauverlaufsAnalyseOptionen(argList);
    }

    /**
     * führt den timergesteuerten Task aus.
     *
     * @see java.util.TimerTask#run()
     */
    @Override
    public synchronized void run() {
        StauverlaufsAnalyse.logger.finest("Zyklischer Task Stauverlaufsanalyse");
        final long now = System.currentTimeMillis();
        for (final IStauobjektBestimmung sobModul : sobModule.values()) {
            final long nextTime = sobModul.getNextExecutionTime();
            if (now >= nextTime) {
                sobModul.setLastExecutionTime(nextTime);
                sobModul.calculate(nextTime);
                final IStauverlaufsPrognose prognoseModul = sobModul.getPrognoseModul();
                if (prognoseModul != null) {
                    for (final IStau stau : sobModul.getStauList()) {
                        stau.setPrognose(prognoseModul.prognose(stau, nextTime));
                        stau.publizierePrognoseDaten();
                    }
                }
            }
        }
    }

    @Override
    public void uncaughtException(final Thread t, final Throwable e) {
        if (StauverlaufsAnalyse.logger != null) {
            String message = MessageFormat.format("Ein Thread {0} wurde unerwartet beendet: {1}", t, e.getMessage());
            if (betriebsmeldungssender != null) {
                betriebsmeldungssender.fehler(null, message);
            }
            Debug.getLogger().error(message);
        } else {
            String message = MessageFormat.format("Ein Thread {0} wurde unerwartet beendet: {1}", t, e);
            System.err.println(message);
            e.printStackTrace();
        }

        System.exit(-1);
    }

    public GanglinienPrognoseClient getGanglinienprognose() {
        return prognose;
    }

    public SvaBetriebsmeldungssender getBetriebsmeldungssender() {
        return betriebsmeldungssender;
    }

	public int getMaxGapSize() {
		return maxGapSize;
	}
	
	public void setMaxGapSize(int size) {
		maxGapSize = size;
	}
}
