/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.progstau;

import java.util.ArrayList;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;

/**
 * Die Daten eines Prognoseschritts.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SvpPrognoseDaten extends ArrayList<SvpSchritt> {

    /**
     * ID für die Serialisierung des Objekts.
     */
    private static final long serialVersionUID = 1L;

    /**
     * die Datenbeschreibung zum Versand der Prognosedaten an den Datenverteiler.
     */
    private DataDescription dataDesc;

    /**
     * Zeitliche Schrittweite der Prognoseergebnisse.
     */
    private long schrittWeite;

    /**
     * Dauer des Staus gemessen vom aktuellen Zeitpunkt.
     */
    private long dauer;

    /**
     * Zeitpunkt der Stauauflösung.
     */
    private long aufloesungsZeit = Long.MAX_VALUE;

    /**
     * Maximale Staulänge.
     */
    private long maxLaenge;

    /**
     * Zeitpunkt der maximalen Staulänge.
     */
    private long maxLaengeZeit = Long.MAX_VALUE;

    /** der Zeitpunkt der Prognose. */
    private final long prognoseZeitPunkt;

    private StauverlaufsAnalyse applikation;

    /**
     * Standardkonstruktor.
     *
     * @param applikation       die Applikation, die die Stauverlaufsanalyse
     *                          durchführt
     * @param prognoseZeitPunkt der Zeitpunkt für den die Prognosedaten erstellt
     *                          werden
     */
    public SvpPrognoseDaten(StauverlaufsAnalyse applikation, final long prognoseZeitPunkt) {
        this.applikation = applikation;
        DataModel model = applikation.getDav().getDataModel();
        dataDesc = new DataDescription(model.getAttributeGroup(SvpDefs.PID_ATG_STAU_VERLAUF),
                model.getAspect(SvpDefs.PID_ASP_PROGNOSE_NORMAL));
        this.prognoseZeitPunkt = prognoseZeitPunkt;
    }

    /**
     * liefert den prognostizierten Auflösungszeitpunkt des Staus.
     *
     * @return den Zeitpunkt
     */
    public final long getAufloesungsZeit() {
        return aufloesungsZeit;
    }

    /**
     * erzeugt einen Datensatz zum Versand der Prognosedaten an den Datenverteiler.
     *
     * @return den erzeugten Datensatz
     */
    public Data getDatenSatz() {
        final Data data = applikation.getDav().createData(dataDesc.getAttributeGroup());

        data.getTimeValue("Schrittweite").setMillis(getSchrittWeite());
        data.getTimeValue("Dauer").setMillis(getDauer());
        data.getTimeValue("AuflösungsZeit").setMillis(getAufloesungsZeit());
        data.getScaledValue("MaxLänge").set(getMaxLaenge());
        data.getTimeValue("MaxLängeZeit").setMillis(getMaxLaengeZeit());

        final Data.Array verlauf = data.getArray("Prognoseverlauf");
        verlauf.setLength(size());

        int idx = 0;
        for (final SvpSchritt schritt : this) {
            verlauf.getItem(idx).getScaledValue("Zufluss").set(schritt.getZufluss());
            final double kapazitaet = schritt.getKapazitaet();
            if ((kapazitaet < 0) || (kapazitaet > 100000000)) {
                verlauf.getItem(idx).getUnscaledValue("Kapazität").setText("nicht ermittelbar");
            } else {
                verlauf.getItem(idx).getScaledValue("Kapazität").set(kapazitaet);
            }
            verlauf.getItem(idx).getScaledValue("Länge").set(schritt.getLaenge());
            verlauf.getItem(idx).getTimeValue("VerlustZeit").setMillis(schritt.getVerlustZeit());
            verlauf.getItem(idx).getScaledValue("vKfz").set(schritt.getVKfz());
            idx++;
        }

        return data;
    }

    /**
     * liefert die Dauer das Staus zurück.
     *
     * @return die Dauer
     */
    public long getDauer() {
        return dauer;
    }

    /**
     * liefert die maximale Staulänge im Prognosezeitraum.
     *
     * @return die Länge
     */
    public final long getMaxLaenge() {
        return maxLaenge;
    }

    /**
     * liefert den Zeitpunkt, an dem die maximale Staulänge im Prognosezeitraum
     * erreicht wurde.
     *
     * @return den Zeitpunkt
     */
    public final long getMaxLaengeZeit() {
        return maxLaengeZeit;
    }

    /**
     * liefert den Zeitpunkt, für den die Prognose erstellt wurde.
     *
     * @return den Zeitpunkt
     */
    public long getPrognoseZeitPunkt() {
        return prognoseZeitPunkt;
    }

    /**
     * liefert die Schrittweite, mit der die Prognose ausgeführt wurde.
     *
     * @return die Schrittweite.
     */
    public final long getSchrittWeite() {
        return schrittWeite;
    }

    /**
     * setzt die Auflösungszeit des Staus. Da keine Werte &lt;= 0 innerhalb dieses
     * Attributes über den Datenverteiler versendet werden, wird die Auflösungszeit
     * beim Setzen eines solchen Wertes auf Long.MAX_VALUE gesetzt.
     *
     * @param value Festzulegender aufloesungsZeit
     */
    public final void setAufloesungsZeit(final long value) {
        if (value <= 0) {
            aufloesungsZeit = Long.MAX_VALUE;
        } else {
            aufloesungsZeit = value;
        }
    }

    /**
     * setzt die Dauer des Staus.
     *
     * @param value Festzulegender dauer
     */
    public final void setDauer(final long value) {
        dauer = value;
    }

    /**
     * setzt die maximale Länge.
     *
     * @param value Festzulegender maxLaenge
     */
    public final void setMaxLaenge(final long value) {
        maxLaenge = value;
    }

    /**
     * setzt den Zeitpunkt, zu dem die maximale Länge zu verzeichnen war.
     *
     * @param value Festzulegender maxLaengeZeit
     */
    public final void setMaxLaengeZeit(final long value) {
        maxLaengeZeit = value;
    }

    /**
     * setzt die Schrittweite der Prognosedaten.
     *
     * @param value Festzulegender schrittWeite
     */
    public final void setSchrittWeite(final long value) {
        schrittWeite = value;
    }

    public DataDescription getDataDesc() {
        return dataDesc;
    }
}
